#include <stdio.h>
#include <string.h>
#include <ogcsys.h>
#include <sdcard/wiisd_io.h>

#include "sdhc.h"

/* IOCTL comamnds */
#define IOCTL_SDHC_INIT		0x01
#define IOCTL_SDHC_READ		0x02
#define IOCTL_SDHC_WRITE	0x03
#define IOCTL_SDHC_ISINSERTED	0x04

#define SDHC_HEAPSIZE		0x800

/* Variables */
static char fs[] ATTRIBUTE_ALIGN(32) = "/dev/sdio/sdhc";

static s32  hid = -1, fd = -1;


bool SDHC_Init(void)
{
	s32 ret;

	/* Already open */
	if (fd > 0)
		return true;

	/* Create heap */
	if (hid < 0) {
		hid = iosCreateHeap(SDHC_HEAPSIZE);
		if (hid < 0)
			goto err;
	}

	/* Open SDHC device */
	fd = IOS_Open(fs, 0);
	if (fd < 0)
		goto err;

	/* Initialize SDHC */
	ret = IOS_IoctlvFormat(hid, fd, IOCTL_SDHC_INIT, ":");
	if (ret)
		goto err;

	return true;

err:
	/* Close SDHC device */
	if (fd > 0) {
		IOS_Close(fd);
		fd = -1;
	}

	return false;
}

bool SDHC_Close(void)
{
	/* Close SDHC device */
	if (fd > 0) {
		IOS_Close(fd);
		fd = -1;
	}

	return true;
}

bool SDHC_IsInserted(void)
{
	s32 ret;

	/* Check if SD card is inserted */
	ret = IOS_IoctlvFormat(hid, fd, IOCTL_SDHC_ISINSERTED, ":");

	return (!ret) ? true : false;
}

bool SDHC_ReadSectors(u32 sector, u32 count, void *buffer)
{
	void *buf = NULL;
	s32   ret;

	/* Device not opened */
	if (fd < 0)
		return false;

	/* Buffer not aligned */
	if ((u32)buffer & 0x1F) {
		u32 cnt;

		/* Allocate memory */
		buf = iosAlloc(hid, SDHC_SECTOR_SIZE);

		/* Read data */
		for (cnt = 0; cnt < count; cnt++) {
			void *ptr = ((u8 *)buffer) + (SDHC_SECTOR_SIZE * cnt);

			/* Read sector */
			ret = IOS_IoctlvFormat(hid, fd, IOCTL_SDHC_READ, "ii:d", sector + cnt, 1, buf, SDHC_SECTOR_SIZE);
			if (ret)
				goto out;

			/* Copy sector */
			memcpy(ptr, buf, SDHC_SECTOR_SIZE);
		}
	} else
		ret = IOS_IoctlvFormat(hid, fd, IOCTL_SDHC_READ, "ii:d", sector, count, buffer, SDHC_SECTOR_SIZE * count);

out:
	/* Free memory */
	if (buf)
		iosFree(hid, buf);

	return (!ret) ? true : false;
}

bool SDHC_WriteSectors(u32 sector, u32 count, void *buffer)
{
	void *buf = NULL;
	s32   ret;

	/* Device not opened */
	if (fd < 0)
		return false;

	/* Buffer not aligned */
	if ((u32)buffer & 0x1F) {
		u32 cnt;

		/* Allocate memory */
		buf = iosAlloc(hid, SDHC_SECTOR_SIZE);

		/* Read data */
		for (cnt = 0; cnt < count; cnt++) {
			void *ptr = ((u8 *)buffer) + (SDHC_SECTOR_SIZE * cnt);

			/* Copy sector */
			memcpy(buf, ptr, SDHC_SECTOR_SIZE);

			/* Write sector */
			ret = IOS_IoctlvFormat(hid, fd, IOCTL_SDHC_WRITE, "ii:d", sector + cnt, 1, buf, SDHC_SECTOR_SIZE);
			if (ret)
				goto out;
		}
	} else
		ret = IOS_IoctlvFormat(hid, fd, IOCTL_SDHC_WRITE, "ii:d", sector, count, buffer, SDHC_SECTOR_SIZE * count);

out:
	/* Free memory */
	if (buf)
		iosFree(hid, buf);

	return (!ret) ? true : false;
}

bool SDHC_ClearStatus(void)
{
	return true;
}


const DISC_INTERFACE __io_sdhc = {
	DEVICE_TYPE_WII_SD,
	FEATURE_MEDIUM_CANREAD | FEATURE_MEDIUM_CANWRITE | FEATURE_WII_SD,
	(FN_MEDIUM_STARTUP)&SDHC_Init,
	(FN_MEDIUM_ISINSERTED)&SDHC_IsInserted,
	(FN_MEDIUM_READSECTORS)&SDHC_ReadSectors,
	(FN_MEDIUM_WRITESECTORS)&SDHC_WriteSectors,
	(FN_MEDIUM_CLEARSTATUS)&SDHC_ClearStatus,
	(FN_MEDIUM_SHUTDOWN)&SDHC_Close
};
